"""
abcplus is a stand-in for the stdlib `abc` module that adds an @finalmethod decorator for final methods
"""

import abc

abstractproperty = abc.abstractproperty  # pylint: disable=invalid-name
abstractmethod = abc.abstractmethod


def finalmethod(funcobj):
    """A decorator indicating final methods.

    Requires that the metaclass is abcplus.ABCMeta or derived from it.  A
    class that has a metaclass derived from ABCMeta cannot be
    instantiated if any of the final methods are overridden.
    The abstract methods can be called using any of the normal
    'super' call mechanisms.

    Usage:

        class C:
            __metaclass__ = abcplus.ABCMeta
            @abcplus.finalmethod
            def my_final_method(self, ...):
                ...
    """
    funcobj.__isfinalmethod__ = True
    return funcobj


class ABCMeta(abc.ABCMeta):
    """A modified version of ABCMeta metaclass that adds support for checking final methods"""
    def __new__(mcls, name, bases, namespace):
        """Check the inheritance tree for any final methods and then verify we aren't overriding them."""

        def _check_base_cls(name, namespace, base):
            """Recursively check parent classes for final methods that we have overridden and throw TypeError if found.
            """
            if base.__bases__:
                for base_parent in base.__bases__:
                    _check_base_cls(name, namespace, base_parent)

            for method in getattr(base, "__finalmethods__", set()):
                if method in namespace:
                    error = "Class {cls} cannot override final method {method} from parent class {base}".format(
                        cls=name,
                        method=method,
                        base=base.__name__
                    )
                    raise TypeError(error)

        for base in bases:
            _check_base_cls(name, namespace, base)

        cls = super(ABCMeta, mcls).__new__(mcls, name, bases, namespace)

        finals = set(ns_name
                     for ns_name, ns_value in namespace.items()
                     if getattr(ns_value, "__isfinalmethod__", False))
        cls.__finalmethods__ = frozenset(finals)

        return cls



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import setuptools

setuptools.setup(
    name="abcplus",
    version="0.1.0",
    author="Greg Hill",
    author_email="jimbobhickville@gmail.com",
    description="An amplified version of abc for Abstract Base Classes plus more",
    url="https://www.github.com/jimbobhickville/abcplus",
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    install_requires=[],
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python"
    ],
    scripts=[
    ],
)
